-- $Id: luatexko-normalize.lua,v 1.5 2012/06/01 08:31:05 nomos Exp $
--
-- Copyright (c) 2013 Dohyun Kim  <nomos at ktug org>
--
-- This work may be distributed and/or modified under the
-- conditions of the LaTeX Project Public License, either version 1.3c
-- of this license or (at your option) any later version.
-- The latest version of this license is in
--   http://www.latex-project.org/lppl.txt
-- and version 1.3c or later is part of all distributions of LaTeX
-- version 2006/05/20 or later.

module('luatexkonormalize', package.seeall)

luatexbase.provides_module({
    name	= "luatexko-normalize",
    version	= 1.0,
    date	= "2013/03/30",
    author	= "Dohyun Kim",
    description	= "Hangul normalization (chiefly for NFC)",
    license	= "LPPL",
})

local cho   = "[\225\132\128-\225\132\146]"
local jung  = "[\225\133\161-\225\133\181]"
local jong  = "[\225\134\168-\225\135\130]"
local ojong = "[\225\135\131-\225\135\191\237\159\139-\237\159\187]"
local compathanja = "[\239\164\128-\239\168\139]"
local chanjatohanja = dofile(kpse.find_file('compathanja2hanja.lua'))
local jamotocjamo = dofile(kpse.find_file('jamo2compatjamo.lua'))

local gsub = unicode.utf8.gsub
local byte = unicode.utf8.byte
local char = unicode.utf8.char

local syllable2jamo = function(l,v,t)
    l, v = byte(l), byte(v)
    local s = (l - 0x1100) * 21
    s = (s + v - 0x1161) * 28
    if t then
	t = byte(t)
	s = s + t - 0x11a7
    end
    s = s + 0xac00
    return char(s)
end

local hanguldecompose = function(buffer)
    buffer = gsub(buffer, "[가-힣]", function(s)
	s = byte(s) - 0xac00
	local cho = s / (21 * 28) + 0x1100
	local jung = (s % (21 * 28)) / 28 + 0x1161
	local jong = s % 28 + 0x11a7
	if jong > 0x11a7 then
	    return char(cho)..char(jung)..char(jong)
	end
	return char(cho)..char(jung)
    end)
    return buffer
end

local function hanjanormalize(c)
    local hanja = chanjatohanja[byte(c)]
    hanja = hanja and char(hanja)
    return hanja
end

local function jamo2cjamocho(c)
    local jamo = jamotocjamo.ccho[byte(c)]
    jamo = jamo and char(jamo)
    return jamo
end

local function jamo2cjamojung(c)
    local jamo = jamotocjamo.cjung[byte(c)]
    jamo = jamo and char(jamo)
    return jamo
end

local hangulcompose = function(buffer)
    buffer = hanguldecompose(buffer)
    buffer = gsub(buffer, "("..cho..")("..jung..")("..jong..")", syllable2jamo)
    buffer = gsub(buffer, "("..cho..")("..jung..ojong..")", "%1\1%2")
    buffer = gsub(buffer, "("..cho..")("..jung..")", syllable2jamo)
    buffer = gsub(buffer, "([\225\132\128-\225\133\153])\225\133\160", jamo2cjamocho)
    buffer = gsub(buffer, "\225\133\159([\225\133\161-\225\134\161])", jamo2cjamojung)
    buffer = gsub(buffer, "\1", "")
    buffer = gsub(buffer, "("..compathanja..")", hanjanormalize)
    return buffer
end

local loaded = false

function compose()
    if loaded then
	luatexbase.remove_from_callback('process_input_buffer',
	'luatexko-hangul-normalize')
    end
    loaded = true
    luatexbase.add_to_callback('process_input_buffer',
    hangulcompose, 'luatexko-hangul-normalize')
end

function decompose()
    if loaded then
	luatexbase.remove_from_callback('process_input_buffer',
	'luatexko-hangul-normalize')
    end
    loaded = true
    luatexbase.add_to_callback('process_input_buffer',
    hanguldecompose, 'luatexko-hangul-normalize')
end

function unload()
    if loaded then
	luatexbase.remove_from_callback('process_input_buffer',
	'luatexko-hangul-normalize')
    end
    loaded = false
end

